//
// CFloatImage
//
// This class implements the image structure that the scripts will manipulate.
// By storing floats instead of 8 bit integers for each channel the scripts have 
// a much higher degree of freedom with how the pixels are generated.
//


#ifndef FLOATIMAGE_H
#define FLOATIMAGE_H

typedef unsigned int UINT;

class CFloatImage
{
public:
	CFloatImage();
	virtual ~CFloatImage();

	CFloatImage &operator=(const CFloatImage &other);

	int AddRef();
	int Release();

	int Create(UINT w, UINT h);
	void Resize(UINT w, UINT h);

	void SetAlpha(UINT x, UINT y, float a);
	float GetAlpha(UINT x, UINT y);
	float *Alpha(UINT x, UINT y);
	float *Red(UINT x, UINT y);
	float *Green(UINT x, UINT y);
	float *Blue(UINT x, UINT y);
	float *Channel(UINT ch, UINT x, UINT y);

	float *pixels;
	UINT   width;
	UINT   height;

private:
	void ResizeWidthAvg(UINT newWidth);
	void ResizeHeightAvg(UINT newHeight);

	int refCount;
};

class asIScriptEngine;
void RegisterScriptImage(asIScriptEngine *engine);

#endif