#include "imagewnd.h"
#include "pngfile.h"
#include "resource.h"
#include <afxres.h>
#include <commdlg.h>	// GetOpenFileName()
#include <stdlib.h>		// _MAX_PATH
#include <cderr.h>		// error codes for CommDlgExtendedError()
#include "pngdlg.h"
#include "floatimage.h"

CImageWnd::CImageWnd()
{
	scale = 1.0;
	viewAlpha = false;
	viewTiled = false;
}

void CImageWnd::CopyImage(CFloatImage *img)
{
	originalImage.Create(img->width, img->height);

	float *fpixels = img->pixels;
	BYTE *pixels = (BYTE*)originalImage.pixels;

	for( UINT n = 0; n < img->width*img->height*4; n++ )
	{
		float f = *fpixels++;
		if( f > 1 ) 
			f = 1;
		if( f < 0 ) 
			f = 0;

		*pixels++ = BYTE(f*255.99999f);
	}

	ProcessImage();
}

int CImageWnd::Create(acCWindow *parent)
{
	HICON hIcon = (HICON)LoadImage(GetModuleHandle(0), MAKEINTRESOURCE(IDI_ANGELCODE), IMAGE_ICON, LR_DEFAULTSIZE, LR_DEFAULTSIZE, LR_SHARED);
	HICON hIconSmall = (HICON)LoadImage(GetModuleHandle(0), MAKEINTRESOURCE(IDI_ANGELCODE), IMAGE_ICON, 16, 16, LR_SHARED);

	int r = RegisterClass("ImageWnd", 0, 0, hIcon, hIconSmall, AC_REGDEFCURSOR);
	if( r < 0 ) return r;

	RECT rc = {0, 0, originalImage.width, originalImage.height};
	AdjustWindowRect(&rc, WS_OVERLAPPEDWINDOW, TRUE);

	r = acCWindow::Create("Image Window", rc.right-rc.left, rc.bottom-rc.top, WS_OVERLAPPEDWINDOW, 0, parent, "ImageWnd");
	if( r < 0 ) return r;

	SetMenu(MAKEINTRESOURCE(IDR_IMG_MENU));

	SetAccelerator(MAKEINTRESOURCE(IDR_IMG_ACCEL));

	ShowWindow(hWnd, SW_SHOW);

	return 0;
}

LRESULT CImageWnd::MsgProc(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch( msg )
	{
	case WM_PAINT:
		OnPaint();
		return 0;

	case WM_SIZE:
		OnSize();
		return 0;

	case WM_HSCROLL:
		OnScroll(SB_HORZ, LOWORD(wParam));
		return 0;

	case WM_VSCROLL:
		OnScroll(SB_VERT, LOWORD(wParam));
		return 0;

	case WM_INITMENUPOPUP:
		OnInitMenuPopup((HMENU)wParam, LOWORD(lParam), HIWORD(lParam));
		return 0;

	case WM_COMMAND:
		switch( LOWORD(wParam) )
		{
		case ID_FILE_CLOSE:
			SendMessage(hWnd, WM_CLOSE, 0, 0);
			return 0;

		case ID_FILE_SAVEAS:
			OnSaveAs();
			return 0;

		case IDC_VIEW_SCALE_1_8:
			OnScale(0.125);
			return 0;

		case IDC_VIEW_SCALE_1_4:
			OnScale(0.25);
			return 0;
			
		case IDC_VIEW_SCALE_1_2:
			OnScale(0.5);
			return 0;

		case IDC_VIEW_SCALE_1_1:
			OnScale(1);
			return 0;

		case IDC_VIEW_SCALE_2_1:
			OnScale(2);
			return 0;

		case IDC_VIEW_SCALE_4_1:
			OnScale(4);
			return 0;
			
		case IDC_VIEW_SCALE_8_1:
			OnScale(8);
			return 0;

		case ID_VIEW_CHANNELS_COLOR:
		case ID_VIEW_CHANNELS_ALPHA:
			viewAlpha = LOWORD(wParam) == ID_VIEW_CHANNELS_ALPHA;
			ProcessImage();
			Draw();
			Invalidate(false);
			return 0;

		case ID_VIEW_TILED:
			viewTiled = !viewTiled;
			Draw();
			Invalidate(false);
			return 0;
		}
	}

	return DefWndProc(msg, wParam, lParam);
}

void CImageWnd::OnPaint()
{
	PAINTSTRUCT ps;
	memset(&ps, 0, sizeof(PAINTSTRUCT));

	HDC dc = BeginPaint(hWnd, &ps);

	RECT rc;
	GetClientRect(hWnd, &rc);
	buffer.CopyToDC(dc, 0, 0, rc.right, rc.bottom);

	EndPaint(hWnd, &ps);
}

void CImageWnd::OnSaveAs()
{
	OPENFILENAME ofn;
	char file[_MAX_PATH] = "";

	memset(&ofn, 0, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hWnd;
	ofn.lpstrFilter = "PNG images\0*.PNG\0";
	ofn.lpstrFile = file;
	ofn.nMaxFile = _MAX_PATH;
	ofn.Flags = OFN_OVERWRITEPROMPT|OFN_HIDEREADONLY|OFN_NONETWORKBUTTON|OFN_PATHMUSTEXIST;
	ofn.lpstrDefExt = "as";

	if( GetSaveFileName(&ofn) != 0 )
	{
		// Get user options
		CPngDlg dlg;
		dlg.DoDialog(this);

		// Set new file path
		if( writePNG(file, &originalImage, dlg.saveChannelsMode, dlg.saveInterlaced) < 0 )
			MessageBox(hWnd, "Failed to save image", "Error", MB_OK);
	}
/*	else
	{
		DWORD dw = CommDlgExtendedError();
		switch( dw )
		{
		case CDERR_DIALOGFAILURE:
			OutputDebugString("CDERR_DIALOGFAILURE\n");
			break;
		case CDERR_FINDRESFAILURE:
			OutputDebugString("CDERR_FINDRESFAILURE\n");
			break;
		case CDERR_INITIALIZATION:
			OutputDebugString("CDERR_INITIALIZATION\n");
			break;
		case CDERR_LOADRESFAILURE:
			OutputDebugString("CDERR_LOADRESFAILURE\n");
			break;
		case CDERR_LOADSTRFAILURE:
			OutputDebugString("CDERR_LOADSTRFAILURE\n");
			break;
		case CDERR_LOCKRESFAILURE:
			OutputDebugString("CDERR_LOCKRESFAILURE\n");
			break;
		case CDERR_MEMALLOCFAILURE:
			OutputDebugString("CDERR_MEMALLOCFAILURE\n");
			break;
		case CDERR_MEMLOCKFAILURE:
			OutputDebugString("CDERR_MEMLOCKFAILURE\n");
			break;
		case CDERR_NOHINSTANCE:
			OutputDebugString("CDERR_NOHINSTANCE\n");
			break;
		case CDERR_NOHOOK:
			OutputDebugString("CDERR_NOHOOK\n");
			break;
		case CDERR_NOTEMPLATE:
			OutputDebugString("CDERR_NOTEMPLATE\n");
			break;
		case CDERR_REGISTERMSGFAIL:
			OutputDebugString("CDERR_REGISTERMSGFAIL\n");
			break;
		case CDERR_STRUCTSIZE:
			OutputDebugString("CDERR_STRUCTSIZE\n");
			break;
		case FNERR_BUFFERTOOSMALL:
			OutputDebugString("FNERR_BUFFERTOOSMALL\n");
			break;
		case FNERR_INVALIDFILENAME:
			OutputDebugString("FNERR_INVALIDFILENAME\n");
			break;
		case FNERR_SUBCLASSFAILURE:
			OutputDebugString("FNERR_SUBCLASSFAILURE\n");
			break;
		}
	}*/
}

void CImageWnd::OnSize()
{
	// Do we need to add scroll bars?
	UpdateScrollBars();

	// Resize back buffer
	RECT rc;
	GetClientRect(hWnd, &rc);
	buffer.Create(rc.right, rc.bottom);
	Draw();

	// Show back buffer
	Invalidate(false);
}

void CImageWnd::UpdateScrollBars()
{
	RECT rc;

	int imgWidth = int(originalImage.width * scale);
	int imgHeight = int(originalImage.height * scale);

	GetClientRect(hWnd, &rc);
	ShowScrollBar(hWnd, SB_HORZ, rc.right < imgWidth);

	GetClientRect(hWnd, &rc);
	ShowScrollBar(hWnd, SB_VERT, rc.bottom < imgHeight);

	// Set scroll bar range
	GetClientRect(hWnd, &rc);
	SCROLLINFO info;
	info.cbSize = sizeof(SCROLLINFO);
	info.fMask = SIF_RANGE | SIF_POS;
	if( imgWidth - rc.right > 0 )
	{
		GetScrollInfo(hWnd, SB_HORZ, &info);

		info.nMin = 0;
		info.nMax = imgWidth - rc.right;
		info.nPos = info.nPos > info.nMax ? info.nMax : info.nPos;
	}
	else
	{
		info.nMin = 0;
		info.nMax = 0;
		info.nPos = 0;
	}
	SetScrollInfo(hWnd, SB_HORZ, &info, TRUE);
	
	if( imgHeight - rc.bottom > 0 )
	{
		GetScrollInfo(hWnd, SB_HORZ, &info);

		info.nMin = 0;
		info.nMax = imgHeight - rc.bottom;
		info.nPos = info.nPos > info.nMax ? info.nMax : info.nPos;
	}
	else
	{
		info.nMin = 0;
		info.nMax = 0;
		info.nPos = 0;
	}
	SetScrollInfo(hWnd, SB_VERT, &info, TRUE);
}

void CImageWnd::Draw()
{
	// Clear background
	for( int n = 0; n < buffer.width * buffer.height; n++ )
	{
		buffer.pixels[n] = 0x555555;
	}

	int imgWidth = int(originalImage.width * scale);
	int imgHeight = int(originalImage.height * scale);

	int x1 = buffer.width/2 - imgWidth/2 - 1;
	int x2 = x1 + imgWidth + 1; 
	int y1 = buffer.height/2 - imgHeight/2 - 1;
	int y2 = y1 + imgHeight + 1; 

	if( x1 < 0 ) x1 = -GetHorzScroll() - 1;
	if( y1 < 0 ) y1 = -GetVertScroll() - 1;

	// Draw vertical lines
	if( x1 >= 0 )
	{
		int min = y1 < 0 ? 0 : y1;
		int max = y2 + 1 > buffer.height ? buffer.height : y2 + 1;
		for( int n = min; n < max; n++ )
		{
			buffer.pixels[x1 + n*buffer.width] = 0;
		}
	}

	if( x2 < buffer.width )
	{
		int min = y1 < 0 ? 0 : y1;
		int max = y2 + 1 > buffer.height ? buffer.height : y2 + 1;
		for( int n = min; n < max; n++ )
		{
			buffer.pixels[x2 + n*buffer.width] = 0;
		}
	}

	// Draw horizontal lines
	if( y1 >= 0 )
	{
		int min = x1 < 0 ? 0 : x1;
		int max = x2 + 1 > buffer.width ? buffer.width : x2 + 1;
		for( int n = min; n < max; n++ )
		{
			buffer.pixels[n + y1*buffer.width] = 0;
		}
	}

	if( y2 < buffer.height )
	{
		int min = x1 < 0 ? 0 : x1;
		int max = x2 + 1 > buffer.width ? buffer.width : x2 + 1;
		for( int n = min; n < max; n++ )
		{
			buffer.pixels[n + y2*buffer.width] = 0;
		}
	}

	// Copy image to the backbuffer
	if( !viewTiled )
		DrawImage(x1+1, y1+1);
	else
	{
		x1++;
		y1++;
		while( x1 > 0 ) x1 -= imgWidth;
		while( y1 > 0 ) y1 -= imgHeight;
		int sx = x1;

		while( y1 < buffer.height )
		{
			x1 = sx;
			while( x1 < buffer.width )
			{
				DrawImage(x1, y1);

				x1 += imgWidth;
			}

			y1 += imgHeight;
		}
	}
}

void CImageWnd::DrawImage(int dx, int dy)
{
	int ssx = 0;
	int sy = 0;

	if( dx < 0 )
	{
		ssx += -dx;
		dx = 0;
	}
	if( dy < 0 )
	{
		sy += -dy;
		dy = 0;
	}

	int imgWidth = int(originalImage.width * scale);
	int imgHeight = int(originalImage.height * scale);

	int y2 = dy + imgHeight - sy;
	int x2 = dx + imgWidth - ssx;
	if( y2 > buffer.height ) y2 = buffer.height;
	if( x2 > buffer.width ) x2 = buffer.width;

	// Copy image to the backbuffer
	if( scale > 1 )
	{
		int iScale = int(scale);
		for( int y = dy; y < y2; y++ )
		{
			int sx = ssx;
			for( int x = dx; x < x2; x++ )
			{
				buffer.pixels[x + y*buffer.width] = image.pixels[sx/iScale + sy/iScale*image.width];
				sx++;
			}
			sy++;
		}
	}
	else
	{
		for( int y = dy; y < y2; y++ )
		{
			int sx = ssx;
			for( int x = dx; x < x2; x++ )
			{
				buffer.pixels[x + y*buffer.width] = image.pixels[sx + sy*image.width];
				sx++;
			}
			sy++;
		}
	}
}

int CImageWnd::GetHorzScroll()
{
	SCROLLINFO info;
	info.cbSize = sizeof(SCROLLINFO);
	info.fMask = SIF_POS;
	if( GetScrollInfo(hWnd, SB_HORZ, &info) )
		return info.nPos;

	return 0;
}

int CImageWnd::GetVertScroll()
{
	SCROLLINFO info;
	info.cbSize = sizeof(SCROLLINFO);
	info.fMask = SIF_POS;
	if( GetScrollInfo(hWnd, SB_VERT, &info) )
		return info.nPos;

	return 0;
}

void CImageWnd::OnScroll(UINT scrollBar, UINT action)
{
	SCROLLINFO info;
	info.cbSize = sizeof(SCROLLINFO);
	info.fMask = SIF_POS | SIF_TRACKPOS;
	if( GetScrollInfo(hWnd, scrollBar, &info) )
	{
		switch( action )
		{
		case SB_PAGEUP:
			info.nPos -= image.width/8 ? image.width/8 : 1;
			break;

		case SB_PAGEDOWN:
			info.nPos += image.width/8 ? image.width/8 : 1;
			break;

		case SB_LINEUP:
			info.nPos -= image.width/64 ? image.width/64 : 1;
			break;

		case SB_LINEDOWN:
			info.nPos += image.width/64 ? image.width/64 : 1;
			break;

		case SB_THUMBTRACK:
		case SB_THUMBPOSITION:
			info.nPos = info.nTrackPos;
			break;
		}

		info.fMask = SIF_POS;
		SetScrollInfo(hWnd, scrollBar, &info, TRUE);
	}

	Draw();
	Invalidate(false);
}

void CImageWnd::OnScale(float newScale)
{
	scale = newScale;

	ProcessImage();

	UpdateScrollBars();

	Draw();
	Invalidate(false);
}

void CImageWnd::OnInitMenuPopup(HMENU hMenu, int pos, BOOL isWindowMenu)
{
	CheckMenuItem(hMenu, IDC_VIEW_SCALE_1_8, MF_BYCOMMAND | (scale == 0.125 ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem(hMenu, IDC_VIEW_SCALE_1_4, MF_BYCOMMAND | (scale == 0.25 ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem(hMenu, IDC_VIEW_SCALE_1_2, MF_BYCOMMAND | (scale == 0.5 ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem(hMenu, IDC_VIEW_SCALE_1_1, MF_BYCOMMAND | (scale == 1 ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem(hMenu, IDC_VIEW_SCALE_2_1, MF_BYCOMMAND | (scale == 2 ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem(hMenu, IDC_VIEW_SCALE_4_1, MF_BYCOMMAND | (scale == 4 ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem(hMenu, IDC_VIEW_SCALE_8_1, MF_BYCOMMAND | (scale == 8 ? MF_CHECKED : MF_UNCHECKED));

	CheckMenuItem(hMenu, ID_VIEW_CHANNELS_COLOR, MF_BYCOMMAND | (!viewAlpha ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem(hMenu, ID_VIEW_CHANNELS_ALPHA, MF_BYCOMMAND | (viewAlpha ? MF_CHECKED : MF_UNCHECKED));

	CheckMenuItem(hMenu, ID_VIEW_TILED, MF_BYCOMMAND | (viewTiled ? MF_CHECKED : MF_UNCHECKED));
}

void CImageWnd::ProcessImage()
{
	CImage tmp;
	CImage *src;
	if( viewAlpha )
	{
		tmp.Create(originalImage.width, originalImage.height);
		for( int c = 0; c < tmp.height*tmp.width; c++ )
		{
			DWORD a = originalImage.pixels[c]>>24;
			tmp.pixels[c] = a | (a<<8) | (a<<16) | (a<<24);
		}
		src = &tmp;
	}
	else
		src = &originalImage;

	if( scale < 1 ) 
	{
		int imgWidth = int(originalImage.width*scale);
		int imgHeight = int(originalImage.height*scale);

		image.Create(imgWidth, imgHeight);

		int count = int(1/scale);

		for( int y = 0; y < imgWidth; y++ )
		{
			for( int x = 0; x < imgHeight; x++ )
			{
				// Get the average for the pixel
				int a = 0, r = 0, g = 0, b = 0;
				for( int yc = 0; yc < count; yc++ )
				{
					for( int xc = 0; xc < count; xc++ )
					{
						BYTE *p = (BYTE*)&src->pixels[xc+x*count + (yc + y*count)*src->width];
						b += p[0];
						g += p[1];
						r += p[2];
						a += p[3];
					}
				}
				a /= count*count;
				r /= count*count;
				g /= count*count;
				b /= count*count;

				BYTE *p = (BYTE*)&image.pixels[x+y*image.width];
				p[0] = b;
				p[1] = g;
				p[2] = r;
				p[3] = a;
			}
		}
	}
	else
	{
		image.Create(src->width, src->height);

		memcpy(image.pixels, src->pixels, 4*src->width*src->height);
	}
}

